#!/usr/bin/env python
# vim: set ff=unix expandtab: shiftwidth=4 fileencoding=utf-8
"""
Intelligent RObotics 2014 - source code skeleton
(c) 2014 Czech Technical University, Faculty of Electrical Engineering, Department of Cybernetics

TASK 1
======

This is the sample source code for task 1. Students are free to change it in order to solve the task.

Other use is allowed under the terms of the New BSD license.
"""
__author__ =  'Martin Pecka <peckama2 fel.cvut.cz>'
__version__=  '2014.1.1'
__license__ = 'BSD'

# Ros libraries
import roslib; roslib.load_manifest('task1')
import rospy;

# Python libs
import sys, os, math

# numpy, scipy and pylab
import numpy as np
from scipy.io import loadmat,savemat
from numpy.linalg import svd, eig

import cv2

class compute_camera_pose:

    def __init__(self):
        '''Load the saved .mat files.'''

        base_path = os.path.dirname(os.path.realpath(__file__))
        kinect_correspondences_3D_data = loadmat(base_path + '/../data/kinect_correspondences_3D.mat')
        thermo_correspondences_2D_data = loadmat(base_path + '/../data/thermo_correspondences_2D.mat')
        kinect_K_data = loadmat(base_path + '/../data/kinect_K.mat')

        # the arrays for storing correspondences - it holds that
        #   kinect_correspondences_3D[i] is a correspondence to thermo_correspondences_2D[i]

        # the 3D points corresponding to the 2D points clicked in kinect image (transpose to get points as column vectors)
        self.kinect_correspondences_3D = np.matrix(kinect_correspondences_3D_data['kinect_correspondences_3D']).T
        # the 2D points clicked in the thermocam image (transpose to get points as column vectors)
        self.thermo_correspondences_2D = np.matrix(thermo_correspondences_2D_data['thermo_correspondences_2D']).T

        self.kinect_K = np.matrix(kinect_K_data['kinect_K'], np.float32)

    def main_loop(self):

        print 'Starting projection matrix computation'

        points_3D_norm, norm_matrix_3D = self.normalize_data(self.kinect_correspondences_3D)
        points_2D_norm, norm_matrix_2D = self.normalize_data(self.thermo_correspondences_2D)

        num_points = points_3D_norm.shape[1]

        # IMPLEMENT
        # compose matrix A
        # solve for P

        P = np.matrix(np.reshape(P, (3, 4)))

        P_denormalized = norm_matrix_2D.I * P * norm_matrix_3D

        # save the projection matrices
        base_path = os.path.dirname(os.path.realpath(__file__))
        savemat(base_path + '/../data/P.mat', {'P': P})
        savemat(base_path + '/../data/P_denormalized.mat', {'P_denormalized': P_denormalized})

        decomp = cv2.decomposeProjectionMatrix(P_denormalized)
        K = decomp[0]; K = K/K[2,2]
        R = decomp[1]
        t = decomp[2]; t = t/t[3]

        print 'P'
        print P
        print 'P_denormalized'
        print P_denormalized
        print 'K'
        print K
        print 'R'
        print R
        print 't'
        print t
        print

        self.compute_reprojection_error(P_denormalized)

    def compute_reprojection_error(self, P):
        '''Compute and print the reprojection error given the projection matrix P'''
        # IMPLEMENT
        # compute the mean reprojection error given the projection matrix P

        print 'Mean reprojection error is', err, 'px'



    def normalize_data(self, data):
        '''Normalize the given data matrix to have zero mean and unit average distance; also return the normalization matrix'''

        dim = data.shape[0] # the number of dimensions of the point, either 2 or 3
        coords_num = data.shape[1] # the number of normalized points
        coords_sum = [0] * dim # creates [0, 0] or [0, 0, 0]

        # IMPLEMENT
        # you get input data, which are arrays of 2D or 3D points (points in columns)
        # you need to shift the data to have zero mean
        # then you need to scale the data in such way, that the coorinates of an average point are [1,1] or [1,1,1]
        # that means, after scaling, the average point has to have the distance from origin equal to sqrt(2), resp. sqrt(3)
        # the normalization process can then be expressed as a 3-by-3 or 4-by-4 matrix working with homogeneous coordinates like this one:
        #[scale    0        -scale*translation_x
        # 0        scale    -scale*translation_y
        # 0        0        1                    ]

        return data_normalized, normalization_matrix

def main(args):
    # initialize and run ROS node
    node = compute_camera_pose()
    rospy.init_node('compute_camera_pose', anonymous=True)

    # run the main loop
    node.main_loop()

    cv2.destroyAllWindows()

if __name__ == '__main__':
    main(sys.argv)
