#!/usr/bin/env python
# vim: set ff=unix expandtab: shiftwidth=4 fileencoding=utf-8
"""
Intelligent RObotics 2014 - source code skeleton
(c) 2014 Czech Technical University, Faculty of Electrical Engineering, Department of Cybernetics

TASK 1
======

This is the sample source code for task 1. Students are free to change it in order to solve the task.

Other use is allowed under the terms of the New BSD license.
"""
__author__ =  'Martin Pecka <peckama2 fel.cvut.cz>'
__version__=  '2014.1.1'
__license__ = 'BSD'

# Ros libraries
import roslib; roslib.load_manifest('task1')
import rospy;

# Python libs
import sys, os

# numpy, scipy and pylab
import numpy as np
from scipy.io import loadmat,savemat
from pylab import imshow,ginput

# OpenCV
import cv2

class create_correspondences:

    def __init__(self):
        '''Load the saved .mat files.'''

        base_path = os.path.dirname(os.path.realpath(__file__))
        thermo_data = loadmat(base_path + '/../data/thermo_images.mat')
        rgb_data = loadmat(base_path + '/../data/rgb_images.mat')
        depth_data = loadmat(base_path + '/../data/depth_images.mat')
        kinect_K_data = loadmat(base_path + '/../data/kinect_K.mat')

        self.depth_images = depth_data['depth_images']
        self.rgb_images = rgb_data['rgb_images']
        self.thermo_images = thermo_data['thermo_images']
        self.kinect_K = np.matrix(kinect_K_data['kinect_K'], np.float32)

        # the arrays for storing correspondences - it holds that
        #   kinect_correspondences_3D[i] is a correspondence to thermo_correspondences_2D[i]

        # the 3D points corresponding to the 2D points clicked in kinect image
        self.kinect_correspondences_3D = []
        # the 2D points clicked in the thermocam image
        self.thermo_correspondences_2D = []

    def main_loop(self):

        print 'Starting correspondences finding'

        clean_shutdown = True # if Ctrl-C is pushed before all the images are recorded, this is set False
        num_points_per_image = 4 # the number of points you want to select in each image
        ginput_timeout = 1000000000000

        for i in range(len(self.depth_images)):
            if rospy.is_shutdown():
                clean_shutdown = False
                break

            print '============================='
            print 'Processing image nr.', i, 'out of', len(self.depth_images)
            print '============================='

            '''
            Kinect processing
            '''

            # open the clickable interface for selecting image points
            imshow(self.depth_images[i])
            # imshow(self.rgb_images[i]) # if you rather want to select kinect points in RGB image
            print 'Please click ', num_points_per_image, ' points in the image'
            kinect_points = ginput(num_points_per_image, ginput_timeout)
            print 'you clicked:', kinect_points

            # convert the x,y,depth data to 3D points
            kinect_points_3D = self.depth_to_point_cloud(self.depth_images[i], kinect_points)

            # sum concatenates the arrays
            self.kinect_correspondences_3D = sum([kinect_points_3D], self.kinect_correspondences_3D)

            '''
            Thermocam processing
            '''

            # open the clickable interface for selecting image points
            imshow(self.thermo_images[i])
            print 'Please click ', num_points_per_image, ' points in the image'
            thermo_points = ginput(num_points_per_image, ginput_timeout)
            print 'you clicked:', thermo_points

            # sum concatenates the arrays
            self.thermo_correspondences_2D = sum([thermo_points], self.thermo_correspondences_2D)

        if not clean_shutdown:
            print 'Shutdown forced, exiting.'
            return

        print 'Correspondences finding done, saving.'

        # now all the images are loaded, so save them to .mat files
        base_path = os.path.dirname(os.path.realpath(__file__))
        savemat(base_path + '/../data/kinect_correspondences_3D.mat', {'kinect_correspondences_3D': self.kinect_correspondences_3D}, do_compression=True)
        savemat(base_path + '/../data/thermo_correspondences_2D.mat', {'thermo_correspondences_2D': self.thermo_correspondences_2D}, do_compression=True)
        
        print 'Correspondences were successfully saved.'

    def depth_to_point_cloud(self, depth_image, subset_coordinates):
        '''Convert x,y,depth information to 3D points'''
        invK = self.kinect_K.I # save the inverted matrix

        result = []
        
        # IMPLEMENT
        # you will have to implement this method by yourself
        # the result array should be a 3-by-n point cloud (set of 3D points in Kinect frame) 
        # be sure to check if you don't process points with depth 0, because 0 means NaN in the Kinect data

        return result

def main(args):
    # initialize and run ROS node
    node = create_correspondences()
    rospy.init_node('create_correspondences', anonymous=True)

    # run the main loop
    node.main_loop()

    cv2.destroyAllWindows()

if __name__ == '__main__':
    main(sys.argv)
